#include <iostream>
using namespace std;


//for debug: prints a board
void print_board(char board[5][5], int n)
{
    for (int i = 0; i < n; i++)
    {
        for (int j = 0; j < n; j++)
            cout << board[i][j];
        cout << endl;
    }
}

//remove the previously placed ship
void remove_ship(int ship_size, char board[5][5], int row, int col, bool horizontal)
{
    if (horizontal)
        for (int i = 0; i < ship_size; i++) board[row][col+i] = '.';
    else
        for (int i = 0; i < ship_size; i++) board[row+i][col] = '.';
}

//attempts to place the ship of specified size on the given board state,
//beginning at row,col and moving right or down
//if horizontal = true, place ship across row, otherwise place down the column
//returns -1 if the ship cannot be placed because of a missed shot or another ship
//otherwise, places the ship and returns the number of hit shots covered by this ship
int place_ship(int ship_size, int n, char board[5][5], char given_board[5][5], int row, int col, bool horizontal)
{
    int hits = 0;       //count any hits we cover with this ship

    //place ship across row
    if (horizontal)
    {
        //make sure ship can be placed here, and count hits encountered
        for (int i = 0; i < ship_size; i++)
        {
            //can't place here, return failed placement
            if (col+i >= n || board[row][col+i] == 'X')
                return -1;
            //if this location is a hit, add to counter
            else if (given_board[row][col+i] == 'O')
                hits++;
        }
        //made it this far, ship fits, finalize it
        for (int i = 0; i < ship_size; i++) board[row][col+i] = 'X';
    }
    //place ship down column
    else
    {
        //make sure ship can be placed here, and count hits encountered
        for (int i = 0; i < ship_size; i++)
        {
            //can't place here, return failed placement
            if (row+i >= n || board[row+i][col] == 'X')
                return -1;
            //if this location is a hit, add to counter
            else if (given_board[row+i][col] == 'O')
                hits++;
        }
        //made it this far, ship fits, finalize it
        for (int i = 0; i < ship_size; i++) board[row+i][col] = 'X';
    }
    
    return hits;        //return number of hits on given board covered by this ship
}

//recursive backtracking to count number of valid layouts for the current board 
int count_layouts(int n, int k, char given_board[5][5], char board[5][5], int ships[5], int hit_count, int hits_covered = 0, int ship_index = 0)
{
    int ship_hits;              //hits for each test ship placement    
    int total_layouts = 0;      //count of total layouts

    //all ships placed, is this a solution?
    if (ship_index == k)
    {
        //if we have accounted for all the hits already on the board, solution!
        if (hits_covered == hit_count)
            return 1;
        //if not - all ships already on the board but not all hits covered - not valid
        return 0;
    }

    //loop board positions, top left to bottom right
    for (int row = 0; row < n; row++)    
    {
        for (int col = 0; col < n; col++)
        {
            //try to place the current ship on the board at current location
            //have to try both horizontal and vertical - assuming ship is more than one space

            //horizontal
            ship_hits = place_ship(ships[ship_index], n, board, given_board, row, col, true);
            //ship fits, recurse
            if (ship_hits != -1)
            {
                total_layouts += count_layouts(n, k, given_board, board, ships, hit_count, hits_covered+ship_hits, ship_index+1);     
                remove_ship(ships[ship_index], board, row, col, true);    //remove ship
            }

            //vertical - only if ship not a single space
            if (ships[ship_index] > 1)
            {                
                ship_hits = place_ship(ships[ship_index], n, board, given_board, row, col, false);
                //ship fits, recurse
                if (ship_hits != -1)
                {
                    total_layouts += count_layouts(n, k, given_board, board, ships, hit_count, hits_covered+ship_hits, ship_index+1);     
                    remove_ship(ships[ship_index], board, row, col, false);    //remove ship
                }
            }
        } //end row loop
    } //end column loop

    return total_layouts;
}

int main()
{
    char given_board[5][5];     //battleship board as given (don't want to overwrite)
    char board[5][5];           //battleship board
    int ships[5];               //ship sizes
    int n, k;

    int hit_count = 0;          //count of hits on given board
    int result;                 //final solution - count of valid layouts

    //read case, copying misses over to a working board
    cin >> n >> k;
    for (int i = 0; i < n; i++)
    {
        for (int j = 0; j < n; j++)
        {
            cin >> given_board[i][j];
            board[i][j] = given_board[i][j];
            if (given_board[i][j] == 'O')
            {
                hit_count++;
                board[i][j] = '.';      //don't copy hits
            }
        }
    }
    for (int i = 0; i < k; i++)
        cin >> ships[i];

    //count valid layouts!
    result = count_layouts(n, k, given_board, board, ships, hit_count);

    cout << result << endl;
}