
import java.io.PrintStream;
import java.util.Arrays;
import java.util.Scanner;

/**
 * Solution to Mini Battleship
 */
public class battleship_vanb
{

    /** The input. */
    private static Scanner sc;

    /** The output. */
    private static PrintStream ps;

    /** The board. */
    private char board[][];

    /** The sizes of the ships. */
    private int ships[];

    /** The size of the board. */
    private int n;

    /**
     * Place one ship.
     * 
     * @param level Which ship to place
     * @return The number of ways to place that ship and all following ships
     */
    private long place( int level )
    {
        long count = 0;
        
        // Have we placed all ships?
        if( level == 0 ) 
        {
            // Then this is 1 way of placing all the ships - UNLESS there are spare Os we didn't cover.
            count = 1;
            for( int i=0; i<n; i++ ) for( int j=0; j<n; j++ ) if( board[i][j]=='O' ) count = 0;
        }
        else
        {
            --level;
            int size = ships[level];
            char save[] = new char[size];

            // Try to place horizontally
            for( int i = 0; i < n; i++ ) for( int j = 0; j <= n - size; j++ )
            {
                // Will it fit starting at board[i][j]?
                boolean fits = true;
                for( int k = 0; k < size; k++ ) if( board[i][j + k] == 'X' || board[i][j + k] == '#' ) fits = false;
                
                if( fits )
                {
                    // Put it on the board, save what's under it
                    for( int k = 0; k < size; k++ )
                    {
                        save[k] = board[i][j + k];
                        board[i][j + k] = '#';
                    }
                    
                    // Place the next ship
                    count += place( level );
                    
                    // Undo - take the ship off the board, restore what was under it
                    for( int k = 0; k < size; k++ ) board[i][j + k] = save[k];
                }
            }

            // Try to place vertically
            if( size>1 ) for( int j = 0; j < n; j++ ) for( int i = 0; i <= n - size; i++ )
            {
                // Will it fit starting at board[i][j]?
                boolean fits = true;
                for( int k = 0; k < size; k++ ) if( board[i + k][j] == 'X' || board[i + k][j] == '#' ) fits = false;
                if( fits )
                {
                    // Put it on the board, save what's under it
                    for( int k = 0; k < size; k++ )
                    {
                        save[k] = board[i + k][j];
                        board[i + k][j] = '#';
                    }
                    
                    // Place the next ship
                    count += place( level );
                    
                    // Undo - take the ship off the board, restore what was under it
                    for( int k = 0; k < size; k++ ) board[i + k][j] = save[k];
                }
            }
        }

        return count;
    }

    /**
     * Do it.
     */
    private void doit()
    {
        n = sc.nextInt();
        int k = sc.nextInt();
        
        // Read the board
        board = new char[n][];
        for( int i=0; i<n; i++ ) board[i] = sc.next().toCharArray();
        
        // Read the ship sizes
        ships = new int[k];
        for( int i=0; i<k; i++ ) ships[i] = sc.nextInt();
        
        // Sort the ships, so we'll try to place the largest one first.
        // This will make things a little faster.
        Arrays.sort( ships );

        // This is it!
        ps.println( place( k ) );
    }

    /**
     * The main method.
     *
     * @param args the arguments (unused)
     */
    public static void main( String[] args )
    {
        sc = new Scanner( System.in );
        ps = System.out;
        new battleship_vanb().doit();
    }

}
